'  ***************************       SOLEX       ****************************
'
'  Numerical Integration for Real Time Production of Fundamental Ephemerides
'
'        by a Fast Extrapolation Method Devised for the Solar System
'
'  ...............    Copyright (C) 1996  Aldo Vitagliano    ...............
'
'  *************************************************************************
'                                WARNING !!!
'   Recompilation and linking of this code with a Microsoft Basic Compiler
'   produces a .EXE file running at a considerably slower speed than the
'   original SOLEX.EXE file (20% to 600% [!] slower depending on the machine).
'   Since speed is crucial for this program, I have provided an utility
'   program (SPEEDSOL.EXE) which can be run to restore the original speed
'   to a modified SOLEX.EXE file. Mail the author for further information.
'                  E-mail  aldov@chemna.dichi.unina.it
'  *************************************************************************
'
'
DECLARE SUB Header ()
DECLARE SUB Integrate (Njump%, N%, T#, X#(), Y#(), Z#(), VX#(), VY#(), VZ#())
DECLARE SUB TToDate (TP#, Y%, MO%, D%, HR%, MIN%, SEC%)
DECLARE SUB ScanLib (NST%, YMIN#, YMAX#, DjLib() AS LONG)
DECLARE SUB GetStart (TJ#, Nday#, NUM%, NST%, DjLib() AS LONG, M0#(), RJ0#(), PLN() AS STRING * 8, X#(), Y#(), Z#(), VX#(), VY#(), VZ#())
DECLARE SUB SetOrder (DT#, NO%)
DECLARE SUB PrintOut (FL%, N%, T#, TC#, PLN() AS STRING * 8, X#(), Y#(), Z#())
DECLARE SUB FixV (N%, DT#, DT0#, VX#(), VY#(), VZ#())
DECLARE SUB FixBaric (N%, M0#(), X#(), Y#(), Z#(), VX#(), VY#(), VZ#(), XC0#, YC0#, ZC0#)
DECLARE SUB RotInv (N%, X#(), Y#(), Z#())
DECLARE SUB RotAx (N%, X#(), Y#(), Z#())
DECLARE SUB PrecAng (DJP#)
DECLARE SUB CAcc (N%, X#(), Y#(), Z#(), AX#(), AY#(), AZ#(), M#())
DECLARE SUB STORM8 (N%, X#(), Y#(), Z#(), VX#(), VY#(), VZ#())
DECLARE FUNCTION Dj2000# (YR#, MM#, DD#)

' ............................ Initialization and Input ....................

DEFINT F, H-L, N: DEFDBL A-E, G, M, O-Z
DIM M0(20), M(20), RJ0(20), X(20), Y(20), Z(20), VX(20), VY(20), VZ(20)
DIM DjLib(4095) AS LONG
DIM PLN(20) AS STRING * 8
DIM AK(7, 8)
' ............... Coefficients of the Extrapolation Series .............
DATA 1, 1, 0, 0, 0, 0, 0, 0, 0
DATA 3, -1, 4, 0, 0, 0, 0, 0, 0
DATA 240, 10, -256, 486, 0, 0, 0, 0, 0
DATA 2520, -7, 896, -6561, 8192, 0, 0, 0, 0
DATA 362880, 42, -24576, 531441, -2097152, 1953125, 0, 0, 0
DATA 19958400, -66, 168960, -9743085, 92274688, -244140625, 181398528, 0, 0
' the following combined coefficients are specific for the Moon orbit
' with n=7 and H=3d
' DATA 922442320800, 33462, -480397632, 71157646989, -1394145099776, 7143798828125, -10686833516736, 5788944826368, 0
' the following are specific for n=7 and H=3.75d
DATA 758788430400,33462,-411126144,59537158083,-1157698420736,5911865234375,-8827555291776,4773050843136,0
'
DATA 49037788800, -26,1153152,-387420489,14394851328,-128173828125,322333846848,-549755813888,390625000000
FOR I = 0 TO 7: FOR J = 0 TO 8: READ AK(I, J): NEXT: NEXT
' ............ adopted sequence of stepsizes
DATA 1, 2, 3, 4, 5, 6, 8, 10
FOR K = 0 TO 7: READ QK(K): NEXT

Pi = 4 * ATN(1)
GRAV0 = .0004980621312# '... Gravitational constant (time unit is 1 d)
VL = 25902.06837#  ' ....... Speed of light (length unit is 1 Gm)
OBL0 = 84381.448# * Pi / 648000#  ' ....... Obliquity at JD 2000.0 in radians
SEQ0 = SIN(OBL0): CEQ0 = COS(OBL0)
PHAS = .040700012# '....................... adopted tidal phase angle
CV = .0864#        '............. converts velocity from Km/s to Gm/d

DATA -5.3412578e-008,6.48446607e-008,-3.30323978e-007,-6.3974422e-009,4.755770e-015,2.64932795e-005
READ QJ0, QJ1, QJ2, QJ3, QJ4, QJ5  '... Quadrupole and related parameters

QLOV = QJ4 / PHAS
NTER = 3: NLUN = 13  '... Earth is body n. 3, Moon is body n. 13
CALL Header
CLS
CLS : LOCATE 10, 30: PRINT "Scanning Library ... ";
CALL ScanLib(NST, YMIN, YMAX, DjLib())

DO

CLS
LOCATE 10, 35: PRINT "1. Continue"
LOCATE 12, 35: PRINT "2. Quit"
M$ = ""
WHILE M$ <> "1" AND M$ <> "2": M$ = INKEY$: WEND
IF M$ = "2" THEN CLS : EXIT DO
OutRange% = 1
WHILE OutRange%
   CLS
   INPUT "Enter Date (YYYY,MM,DD)    ", YY, MM, DD
   IF YY >= YMIN AND YY <= YMAX THEN OutRange% = 0
   IF OutRange% THEN
      SOUND 900, 4: CLS
      PRINT "Allowed Range is  "; YMIN; " to "; YMAX
      SLEEP: M$ = INKEY$
   END IF
WEND
   INPUT "Enter Tab. Interval (d)    ", TabStep
   INPUT "Enter # of Output Records  ", Nout

IF Nout <= 0 THEN Nout = 1
IF TabStep = 0 THEN TabStep = 1
D% = SGN(TabStep): TabStep = ABS(TabStep)
TJ = Dj2000(YY, MM, DD) - .5
CALL GetStart(TJ, Nday#, N, NST, DjLib(), M0(), RJ0(), PLN(), X(), Y(), Z(), VX(), VY(), VZ())
RJ1 = 6 * M0(0) * GRAV0 / (VL * VL)
MS% = QK(0): DT0 = 1
T = TJ + .5
' ..........................................................................
S% = 1
IF Nday# < 0 THEN S% = -1  ' ......  S% is negative for backwards integration
DFSTEP = 3.75                   ' ......  Default Optimal Stepsize
DT = DFSTEP: NO = 6             ' ......  Initial Stepsize and Number of stages
Nday# = ABS(Nday#)
Njump = INT(Nday# / DFSTEP)     ' ......  Number of steps to reach target epoch
SStep = Nday# - DFSTEP * Njump  ' ......  Size of last step to reach target epoch
LOCATE 10, 25: PRINT "Jumping to Epoch ";
PRINT USING "#####/##/##"; YY; MM; DD;
COLOR 23: PRINT " ...";
COLOR 7: TC = TIMER

FOR I = 0 TO N
   VX(I) = S% * VX(I) * CV ' ......  Velocity is converted to Gm/d
   VY(I) = S% * VY(I) * CV
   VZ(I) = S% * VZ(I) * CV
NEXT
'.......................................... Baricenter is set to 0,0,0
CALL FixBaric(N, M0(), X(), Y(), Z(), VX(), VY(), VZ(), XC0, YC0, ZC0)
CALL FixV(N, DT, DT0, VX(), VY(), VZ()) '.. Time unit is set equal to stepsize
DJP = T + 30 * S% - S% / 4096  '.... Time at which reference frame is adjusted
CALL PrecAng(DJP)              '.... Precession angles are computed
'                               .... and frame is rotated accordingly
CALL RotAx(N, X(), Y(), Z()): CALL RotAx(N, VX(), VY(), VZ())
'   .................... Integration up to one step before target epoch
CALL Integrate(Njump, N, T, X(), Y(), Z(), VX(), VY(), VZ())
     
IF SStep THEN  ' ............................ Last step to target epoch
   DT0 = DT: DT = SStep: NO = 7
   CALL FixV(N, DT, DT0, VX(), VY(), VZ())
   CALL STORM8(N, X(), Y(), Z(), VX(), VY(), VZ()) ' ... Integration Routine
   T = T + MS% * S% * DT
END IF

D1 = TabStep / 3: ND = INT(D1)
IF ND < D1 THEN ND = ND + 1
DT0 = DT
DT = TabStep / ND  ' ....... Optimal stepsize for given tabulation interval
CALL SetOrder(DT, NO) '..... Optimal order (no. of stages) for given stepsize
CALL FixV(N, DT, DT0, VX(), VY(), VZ()) '... Time unit scaled to stepsize

P% = S% * D%
IF P% < 0 THEN ' ...... Runs backwards if Tabulation interval is negative
   S% = D%
   FOR I = 0 TO N: VX(I) = -VX(I): VY(I) = -VY(I): VZ(I) = -VZ(I): NEXT
END IF
CLS
PRINT "Heliocentric Equatorial Rectangular Coordinates (J2000.0)";
LOCATE 3, 23: PRINT "X (AU)"; TAB(39); "Y (AU)"; TAB(55); "Z (AU)";

FOR JP = 1 TO Nout ' ............. Ephemerides are computed and tabulated
   CALL PrintOut(FLEXIT, N, T, TC, PLN(), X(), Y(), Z())
   IF FLEXIT THEN EXIT FOR
   CALL Integrate(ND, N, T, X(), Y(), Z(), VX(), VY(), VZ())
NEXT JP

LOOP
END

'
'..... Computes all the Accelerations in the system of bodies 0 to N
'
SUB CAcc (N, X(), Y(), Z(), AX(), AY(), AZ(), M()) STATIC
     SHARED QJ0, QJ1, QJ2, QJ3, QJ4, QJ5, QLOV, RJ0(), RJ1, NTER, NLUN, SEQ, CEQ
     FOR I = 0 TO N: AX(I) = 0: AY(I) = 0: AZ(I) = 0: NEXT
' ..... Lunar Geocentric coordinates are stored and converted to baricentric
     XL = X(NLUN): YL = Y(NLUN): ZL = Z(NLUN): X(NLUN) = XL + X(NTER): Y(NLUN) = YL + Y(NTER): Z(NLUN) = ZL + Z(NTER)
     AX(0) = 0: AY(0) = 0: AZ(0) = 0
     FOR J = 1 TO N  ' ...... Computes accelerations by the Sun
        DX = X(J) - X(0): DY = Y(J) - Y(0): DZ = Z(J) - Z(0)
        R2 = DX * DX + DY * DY + DZ * DZ: SR3 = R2 * SQR(R2)
        AFX = (1 + RJ0(J) + RJ1 * R2 / SR3) / SR3 '. Relativistic perturbation
        IF J = NTER THEN '... If Planet is Earth then Oblateness is considered
           RZR = DZ * DZ / R2: RT = SR3 / R2
           AFZ = AFX + QJ2 * (RZR - 3 / 5) / (R2 * SR3): AFX = AFX + QJ2 * (RZR - 1 / 5) / (R2 * SR3)
           ARJ = AFX * M(J): ARJZ = AFZ * M(J): ARI = -AFX * M(0): ARIZ = -AFZ * M(0)
        ELSE
           ARJ = AFX * M(J): ARJZ = ARJ: ARI = -AFX * M(0): ARIZ = ARI
        END IF
        AX(J) = DX * ARI: AY(J) = DY * ARI: AZ(J) = DZ * ARIZ
        AX(0) = AX(0) + DX * ARJ: AY(0) = AY(0) + DY * ARJ: AZ(0) = AZ(0) + DZ * ARJZ
     NEXT J
     RL2 = R2 '.......... Sun-Moon r^2 is stored (Moon must be the last body)
     FOR I = 1 TO N - 1
        FOR J = I + 1 TO N
           IF I = NTER AND J = NLUN THEN '........ Earth-Moon acceleration
              R2 = XL * XL + YL * YL + ZL * ZL: SR3 = R2 * SQR(R2): ARJ = M(J) / SR3: ARI = -M(I) / SR3
              RZR = ZL * ZL / R2
              DZ1 = ZL * CEQ - YL * SEQ: RZR1 = DZ1 * DZ1 / R2
              AFX = 1 + QJ0 + (QJ1 + QJ2 * RZR + QJ3 * RZR1 + QJ5 / RL2) / R2
              AFY = AFX * YL + (2 * QJ3 * DZ1 * SEQ / 5 + (QLOV * YL - QJ4 * XL) / SR3) / R2
              AFZ = AFX * ZL - 2 * (QJ2 * ZL + QJ3 * DZ1 * CEQ + QLOV * ZL * SR3) / (5 * R2)
              AFX = AFX * XL + (QLOV * XL + QJ4 * YL) / (R2 * SR3)
              AX(J) = AX(J) + ARI * AFX: AX(I) = AX(I) + ARJ * AFX
              AY(J) = AY(J) + ARI * AFY: AY(I) = AY(I) + ARJ * AFY
              AZ(J) = AZ(J) + ARI * AFZ: AZ(I) = AZ(I) + ARJ * AFZ
           ELSE     '......................... Planet-Planet accelerations
              DX = X(J) - X(I): DY = Y(J) - Y(I): DZ = Z(J) - Z(I)
              R2 = DX * DX + DY * DY + DZ * DZ: SR3 = R2 * SQR(R2): ARJ = M(J) / SR3: ARI = -M(I) / SR3
              AX(J) = AX(J) + DX * ARI: AY(J) = AY(J) + DY * ARI: AZ(J) = AZ(J) + DZ * ARI
              AX(I) = AX(I) + DX * ARJ: AY(I) = AY(I) + DY * ARJ: AZ(I) = AZ(I) + DZ * ARJ
           END IF
        NEXT J
     NEXT I
     '.............................  Restores Lunar Geocentric Coordinates
     '......................... and computes Lunar Geocentric Acceleration
     X(NLUN) = XL: Y(NLUN) = YL: Z(NLUN) = ZL
     AX(NLUN) = AX(NLUN) - AX(NTER): AY(NLUN) = AY(NLUN) - AY(NTER): AZ(NLUN) = AZ(NLUN) - AZ(NTER)
END SUB

DEFDBL F, L
'.... Computes Julian Day from Date (YR, MN, DY)
'.... Dj2000 is no. of Julian Days from 2000, January 1, 12h TDT (2000.0)
FUNCTION Dj2000 (YR, MN, DY)

    IF MN = 0 THEN MN = 1
    IF DY = 0 THEN DY = 1

    MN1 = MN: YR1 = YR: B = 0
    IF MN < 3 THEN MN1 = MN + 12: YR1 = YR1 - 1
    FLJUL = (YR < 1582) OR (YR = 1582 AND MN < 10) OR (YR = 1582 AND MN = 10 AND DY < 15)
    IF FLJUL = 0 THEN A = INT(YR1 / 100): B = 2 - A + INT(A / 4)
    Dj1 = (INT(365.25 * (YR1 + 4716)) + INT(30.6001# * (MN1 + 1)) + B - 1524.5 - 2451545)
    Dj2000 = Dj1 + DY

END FUNCTION

DEFINT F, L
SUB FixBaric (N, M0(), X(), Y(), Z(), VX(), VY(), VZ(), XC0, YC0, ZC0)
     SHARED NTER, NLUN
     SM = 0: SMX = 0: SMY = 0: SMZ = 0: SVX = 0: SVY = 0: SVZ = 0
     FOR I = 0 TO N
        SM = SM + M0(I)
        IF I = NLUN THEN
           XL = X(I): YL = Y(I): ZL = Z(I): VXL = VX(I): VYL = VY(I): VZL = VZ(I)
           X(I) = XL + X(NTER): Y(I) = YL + Y(NTER): Z(I) = ZL + Z(NTER): VX(I) = VXL + VX(NTER): VY(I) = VYL + VY(NTER): VZ(I) = VZL + VZ(NTER)
        END IF
        SMX = SMX + X(I) * M0(I): SMY = SMY + Y(I) * M0(I): SMZ = SMZ + Z(I) * M0(I): SVX = SVX + VX(I) * M0(I): SVY = SVY + VY(I) * M0(I): SVZ = SVZ + VZ(I) * M0(I)
     NEXT I
     IF SM = 0 THEN SM = 1
     XC0 = SMX / SM: YC0 = SMY / SM: ZC0 = SMZ / SM: VCX = SVX / SM: VCY = SVY / SM: VCZ = SVZ / SM
     FOR I = 0 TO N
        IF I = NLUN THEN
           X(I) = XL: Y(I) = YL: Z(I) = ZL: VX(I) = VXL: VY(I) = VYL: VZ(I) = VZL
        ELSE
           VX(I) = (VX(I) - VCX): VY(I) = (VY(I) - VCY): VZ(I) = (VZ(I) - VCZ)
        END IF
     NEXT I

END SUB

SUB FixV (N, DT, DT0, VX(), VY(), VZ()) STATIC
    SHARED CV, GRAV0, GRAV, M(), M0()
    GRAV = GRAV0 * DT * DT
    FOR I = 0 TO N
       M(I) = M0(I) * GRAV
       VX(I) = VX(I) * DT / DT0
       VY(I) = VY(I) * DT / DT0
       VZ(I) = VZ(I) * DT / DT0
    NEXT

END SUB

'
'... This subroutine reads from the file PLANS.BIN the starting conditions
'... closest in time to the target epoch and returns the number of days
'... needed to reach the target epoch.
'... Within PLANS.BIN, starting positions and velocities are stored in
'... compressed form. Each coordinate is stored in a total of 6 bytes
'... (one INT + one LONG)
'... Scale factors are used to convert them to their final values.
'
SUB GetStart (TJ, Nday#, NUM, NST, DjLib() AS LONG, M0(), RJ0(), PLN() AS STRING * 8, X(), Y(), Z(), VX(), VY(), VZ())
    DIM XX(2, 3)
    CONST maxlong = 4294967295#
    SSL$ = "PLANS.BIN"
    PLD$ = "PLANS.DAT"
    OPEN SSL$ FOR BINARY AS #4
    GET #4, , NUM: LBLOCK = NUM * 36 + 4
    T = 999999999: P& = 3
    CLOSE #4
    FOR J = 1 TO NST
       DJ& = DjLib(J)
       IF TJ = DJ& THEN PF& = P&: T = TJ: EXIT FOR
       IF ABS(TJ - DJ&) < ABS(TJ - T) THEN T = CDBL(DJ&): PF& = P& '...T is nearest day
       P& = P& + LBLOCK
    NEXT J
    Nday# = TJ - T
    N$ = SSL$
    OPEN PLD$ FOR INPUT AS #1
    OPEN N$ FOR BINARY AS #2: SEEK #2, PF&
    GET #2, , DJ&: TJ = DJ&
    INPUT #1, M0(0), RJ0(0), PLN(0)
    X(0) = 0: Y(0) = 0: Z(0) = 0: VX(0) = 0: VY(0) = 0: VZ(0) = 0
    FOR I = 1 TO NUM
       INPUT #1, M0(I), RJ0(I), SF, SFV, PLN(I)
       FOR K = 1 TO 2
          FOR J = 1 TO 3
             GET #2, , Xint%: GET #2, , Xlong&
             IF Xlong& < 0 THEN xfr = Xlong& + maxlong ELSE xfr = Xlong&
             XX(K, J) = (Xint% + xfr / maxlong) / SF
          NEXT J
          SF = SFV
       NEXT K
       X(I) = XX(1, 1): Y(I) = XX(1, 2): Z(I) = XX(1, 3)
       VX(I) = XX(2, 1): VY(I) = XX(2, 2): VZ(I) = XX(2, 3)
    NEXT I
    CLOSE
    EXIT SUB

END SUB

SUB Header
   CLS : V$ = ""
   PRINT : PRINT TAB(20); "*****************************************"
   PRINT TAB(20); "*                 SOLEX                 *";
   PRINT TAB(20); "*****************************************"
   LOCATE 7, 6
   PRINT "  a Program demonstrating the Real Time Computation of Fundamental Ephemerides"
   PRINT : PRINT TAB(39); "by a": PRINT
   PRINT TAB(12); "Fast Extrapolation Method for Integrating the Solar System"
   LOCATE 17, 23: PRINT "Copyright (C) 1996  Aldo Vitagliano"
   PRINT TAB(26); "(aldov@chemna.dichi.unina.it)"
   LOCATE 22, 28: PRINT "Press any key to continue ... ";
   WHILE V$ = "": V$ = INKEY$: WEND
   
END SUB

'........... Adjusts reference frame if time is right for
'........... and calls the integration routine STORM8
SUB Integrate (Njump, N, T, X(), Y(), Z(), VX(), VY(), VZ())
     SHARED MS%, S%, DT, DJP

     FOR JJ = 1 TO Njump
        IF ABS(T - DJP) > 30 THEN
           DJP = DJP + 60 * S%
           CALL RotInv(N, X(), Y(), Z()): CALL RotInv(N, VX(), VY(), VZ())
           CALL PrecAng(DJP)
           CALL RotAx(N, X(), Y(), Z()): CALL RotAx(N, VX(), VY(), VZ())
        END IF
        CALL STORM8(N, X(), Y(), Z(), VX(), VY(), VZ())
        T = T + MS% * S% * DT
     NEXT JJ

END SUB

'................... Computer Precession angles according to Laskar
'................... and the main Nutation term
SUB PrecAng (DJP)
  SHARED Pi, SEQ, CEQ, SEI, CEI, SPA, CPA, SEN, CEN
  MJ = DJP / 365250#: SJ = MJ * 10
  ome = (125.0445222# + SJ * (-1934.1362608# + SJ * (.00207833# + SJ * 2.22E-06))) * Pi / 180
  come = COS(ome)
  DEPS = 9.2 * come * Pi / 648000 '... Nutation change in Obliquity

  PA = (((((-.000000000866# * MJ - .00000004759#) * MJ + .0000002424#) * MJ + .000013095#) * MJ + .00017451#) * MJ - .0018055#) * MJ - .235316#
  PA = (((PA * MJ + .07732#) * MJ + 111.1971) * MJ + 50290.966#) * MJ * Pi / 648000#
  EN = ((((6.6402D-16 * MJ - 2.69151D-15) * MJ - 1.547021D-12) * MJ + 7.521313D-12) * MJ + 6.3190131D-10) * MJ - 3.48388152D-09
  EN = ((((EN * MJ - 1.813065896D-07) * MJ + 2.75036225D-08) * MJ + .000074394531426#) * MJ - .042078604317#) * MJ + 3.052112654975#
  EI = (((((1.2147D-16 * MJ + 7.3759D-17) * MJ - 8.26287D-14) * MJ + 2.50341D-13) * MJ + 2.4650839D-11) * MJ - 5.4000441D-11) * MJ + 1.32115526D-09
  EI = (((EI * MJ - 5.998737027D-07) * MJ - .000016242797091#) * MJ + .002278495537#) * MJ
  EQ0 = ((((.000000000245# * MJ + .00000000579#) * MJ + .0000002787#) * MJ + .000000712#) * MJ - .00003905#) * MJ - .0024967#
  EQANG = (((((EQ0 * MJ - .005138#) * MJ + 1.99925) * MJ - .0155) * MJ - 468.093) * MJ + 84381.448#) * Pi / 648000# + DEPS
  PAS = PA + EN
  SEQ = SIN(EQANG): CEQ = COS(EQANG): SEI = SIN(EI): CEI = COS(EI)
  SPA = SIN(PAS): CPA = COS(PAS): SEN = SIN(EN): CEN = COS(EN)

END SUB

SUB PrintOut (FLEXIT, N, T, TC, PLN() AS STRING * 8, X(), Y(), Z())
   DIM XO(20), YO(20), ZO(20)
   CONST L$ = "   ###.#########"
   CONST S$ = "   ###.#######  "
   'CONST FM$ = "     ###.#########   ###.#########   ###.#########"
   CONST AU = 149.59787066# '... Astronomic Unit (Gm)
   CALL TToDate(T + .5, A%, M%, D%, H%, MIN%, SEC%)
   FOR I = 0 TO N: XO(I) = X(I): YO(I) = Y(I): ZO(I) = Z(I): NEXT
   CALL RotInv(N, XO(), YO(), ZO())
   X0 = XO(0): Y0 = YO(0): Z0 = ZO(0)
   LOCATE 1, 64: PRINT USING "Date:#####/##/##"; A%; M%; D%;
   LOCATE 2, 64: PRINT USING "TDT :   ##:##:##"; H%; MIN%; SEC%;
   LOCATE 3, 64: PRINT USING "Jump time ###.#s"; TIMER - TC
   FOR I = 1 TO N - 1
      IF I < 10 THEN FM$ = L$: K = I ELSE FM$ = S$: K = I + 1
      X = XO(I) - X0: Y = YO(I) - Y0: Z = ZO(I) - Z0
      LOCATE K + 3, 1
      PRINT USING "##  "; I;
      PRINT PLN(I); "  ";
      PRINT USING FM$; X / AU; Y / AU; Z / AU;
   NEXT
   LOCATE I + 5, 1
   PRINT "Geocentric Coordinates"; TAB(24); "(Gm)"; TAB(40); "(Gm)"; TAB(56); "(Gm)"
   LOCATE I + 7, 1
   PRINT USING "##  "; I;
   PRINT PLN(I); "  ";
   PRINT USING L$; XO(I); YO(I); ZO(I);
   LOCATE I + 10, 24
   PRINT "Press a key to continue ... ";
   V$ = "": FLEXIT = 0
   WHILE V$ = ""
      V$ = INKEY$
      IF V$ = CHR$(27) THEN FLEXIT = -1
   WEND
   TC = TIMER
END SUB

' ...... Rotates coordinates to account for precession from J2000 to date
' ...... Five successive rotations are performed, according to
' ...... Bretagnon & Francou, (1988) Astron. Astrophys, 202, 309-315.
'
SUB RotAx (N, X(), Y(), Z())
    SHARED SEQ0, CEQ0, SEQ, CEQ, SEI, CEI, SEN, CEN, SPA, CPA
       FOR I = 0 TO N
          Y = Y(I)
          Y(I) = Y * CEQ0 + Z(I) * SEQ0: Z(I) = Z(I) * CEQ0 - Y * SEQ0
          X = X(I)
          X(I) = X * CEN + Y(I) * SEN: Y(I) = Y(I) * CEN - X * SEN
          Y = Y(I)
          Y(I) = Y * CEI + Z(I) * SEI: Z(I) = Z(I) * CEI - Y * SEI
          X = X(I)
          X(I) = X * CPA - Y(I) * SPA: Y(I) = Y(I) * CPA + X * SPA
          Y = Y(I)
          Y(I) = Y * CEQ - Z(I) * SEQ: Z(I) = Z(I) * CEQ + Y * SEQ
       NEXT I
END SUB

'
' ............ Rotates back the coordinates from date to J2000
'
SUB RotInv (N, X(), Y(), Z())
    SHARED SEQ0, CEQ0, SEQ, CEQ, SEI, CEI, SEN, CEN, SPA, CPA
       FOR I = 0 TO N
          Y = Y(I)
          Y(I) = Y * CEQ + Z(I) * SEQ: Z(I) = Z(I) * CEQ - Y * SEQ
          X = X(I)
          X(I) = X * CPA + Y(I) * SPA: Y(I) = Y(I) * CPA - X * SPA
          Y = Y(I)
          Y(I) = Y * CEI - Z(I) * SEI: Z(I) = Z(I) * CEI + Y * SEI
          X = X(I)
          X(I) = X * CEN - Y(I) * SEN: Y(I) = Y(I) * CEN + X * SEN
          Y = Y(I)
          Y(I) = Y * CEQ0 - Z(I) * SEQ0: Z(I) = Z(I) * CEQ0 + Y * SEQ0
       NEXT I

END SUB

'
'... This subroutine scans the Library file PLANS.BIN, storing the Julian day
'... of each set of starting conditions (measured from J2000) in the array
'... DjLib() and returning the upper and lower date boundary of the Library
'
SUB ScanLib (NST, YMIN, YMAX, DjLib() AS LONG)
    SSL$ = "PLANS.BIN"
    OPEN SSL$ FOR INPUT AS #3: CLOSE #3
    OPEN SSL$ FOR BINARY AS #4
    GET #4, , NUM: LBLOCK = NUM * 36 + 4
    T = 999999999: P& = 3: J = 0: DMIN = 10000000: DMAX = -10000000
    DO WHILE NOT EOF(4)
       J = J + 1
       GET #4, P&, DJ&: IF EOF(4) THEN EXIT DO
       DjLib(J) = DJ&
       IF DJ& < DMIN THEN DMIN = DJ&
       IF DJ& > DMAX THEN DMAX = DJ&
       P& = P& + LBLOCK
    LOOP
    NST = J - 1
    YMIN = INT(DMIN / 365.25 + 1996): YMAX = INT(DMAX / 365.25 + 2004)
    CLOSE #4

END SUB

'..... Select optimal order (no. of stages in STORM8) according to stepsize
SUB SetOrder (DT, NO)
     SELECT CASE DT
        CASE IS > 1.5: NO = 7
        CASE IS > 1: NO = 6
        CASE IS > .5: NO = 5
        CASE IS > .25: NO = 4
        CASE IS > .125: NO = 3
        CASE IS > .03125: NO = 2
        CASE ELSE: NO = 1
     END SELECT

END SUB

'
'. Integration partly based on Extrapol. method for second order conservative
'. conservative equations (Numerical Recipes 2 Ed Cambridge Univ. Press, 1992)
'
SUB STORM8 (N, X(), Y(), Z(), VX(), VY(), VZ()) STATIC
     SHARED AK(), M(), M0(), GRAV, MS%, NO, QK(), CV, DT
     DIM XN(20), YN(20), ZN(20), DELTX(20), DELTY(20), DELTZ(20)
     DIM AX0(20), AY0(20), AZ0(20), AX(20), AY(20), AZ(20)
     DIM X1(7, 20), Y1(7, 20), Z1(7, 20), VX1(7, 20), VY1(7, 20), VZ1(7, 20)
     GR0 = GRAV
     CALL CAcc(N, X(), Y(), Z(), AX0(), AY0(), AZ0(), M()) '... Accelerations
     '......................... Computes NO+1 trial values of the variables
     FOR L = 0 TO NO '.......... NO+1 is the number of stages of the method
         ev = QK(L) / MS%: NSTEP = QK(L) - 1: ev2 = 2 * ev
         GRAV = GR0 / (ev * ev) '... Time unit is scaled so that always h=1
         FOR I = 0 TO N
            M(I) = M0(I) * GRAV
            DELTX(I) = (VX(I) + AX0(I) / ev2) / ev: DELTY(I) = (VY(I) + AY0(I) / ev2) / ev: DELTZ(I) = (VZ(I) + AZ0(I) / ev2) / ev
            XN(I) = X(I) + DELTX(I): YN(I) = Y(I) + DELTY(I): ZN(I) = Z(I) + DELTZ(I)
         NEXT I
         FOR K = 1 TO NSTEP
            CALL CAcc(N, XN(), YN(), ZN(), AX(), AY(), AZ(), M())
            FOR I = 0 TO N
               DELTX(I) = DELTX(I) + AX(I): DELTY(I) = DELTY(I) + AY(I): DELTZ(I) = DELTZ(I) + AZ(I)
               XN(I) = XN(I) + DELTX(I): YN(I) = YN(I) + DELTY(I): ZN(I) = ZN(I) + DELTZ(I)
            NEXT I
         NEXT K
         CALL CAcc(N, XN(), YN(), ZN(), AX(), AY(), AZ(), M())
         FOR I = 0 TO N '....... Stores current trial values of the variables
            X1(L, I) = XN(I): Y1(L, I) = YN(I): Z1(L, I) = ZN(I)
            VX1(L, I) = (DELTX(I) + AX(I) / 2) * ev: VY1(L, I) = (DELTY(I) + AY(I) / 2) * ev: VZ1(L, I) = (DELTZ(I) + AZ(I) / 2) * ev
         NEXT I
     NEXT L
     FOR I = 0 TO N '....... Performs the Extrapolation to h=0 ...............
     X(I) = (AK(NO, 8) * X1(7, I) + AK(NO, 7) * X1(6, I) + AK(NO, 6) * X1(5, I) + AK(NO, 5) * X1(4, I) + AK(NO, 4) * X1(3, I) + AK(NO, 3) * X1(2, I) + AK(NO, 2) * X1(1, I) + AK(NO, 1) * X1(0, I)) / AK(NO, 0)
     Y(I) = (AK(NO, 8) * Y1(7, I) + AK(NO, 7) * Y1(6, I) + AK(NO, 6) * Y1(5, I) + AK(NO, 5) * Y1(4, I) + AK(NO, 4) * Y1(3, I) + AK(NO, 3) * Y1(2, I) + AK(NO, 2) * Y1(1, I) + AK(NO, 1) * Y1(0, I)) / AK(NO, 0)
     Z(I) = (AK(NO, 8) * Z1(7, I) + AK(NO, 7) * Z1(6, I) + AK(NO, 6) * Z1(5, I) + AK(NO, 5) * Z1(4, I) + AK(NO, 4) * Z1(3, I) + AK(NO, 3) * Z1(2, I) + AK(NO, 2) * Z1(1, I) + AK(NO, 1) * Z1(0, I)) / AK(NO, 0)
     VX(I) = (AK(NO, 8) * VX1(7, I) + AK(NO, 7) * VX1(6, I) + AK(NO, 6) * VX1(5, I) + AK(NO, 5) * VX1(4, I) + AK(NO, 4) * VX1(3, I) + AK(NO, 3) * VX1(2, I) + AK(NO, 2) * VX1(1, I) + AK(NO, 1) * VX1(0, I)) / AK(NO, 0)
     VY(I) = (AK(NO, 8) * VY1(7, I) + AK(NO, 7) * VY1(6, I) + AK(NO, 6) * VY1(5, I) + AK(NO, 5) * VY1(4, I) + AK(NO, 4) * VY1(3, I) + AK(NO, 3) * VY1(2, I) + AK(NO, 2) * VY1(1, I) + AK(NO, 1) * VY1(0, I)) / AK(NO, 0)
     VZ(I) = (AK(NO, 8) * VZ1(7, I) + AK(NO, 7) * VZ1(6, I) + AK(NO, 6) * VZ1(5, I) + AK(NO, 5) * VZ1(4, I) + AK(NO, 4) * VZ1(3, I) + AK(NO, 3) * VZ1(2, I) + AK(NO, 2) * VZ1(1, I) + AK(NO, 1) * VZ1(0, I)) / AK(NO, 0)
     NEXT
     GRAV = GR0 '.......... Restores MG values for time unit = stepsize H
     FOR I = 0 TO N: M(I) = M0(I) * GRAV: NEXT
END SUB

SUB TToDate (TP#, Y%, MO%, D%, HR%, MIN%, SEC%)
     T# = TP# + 2451545#
     GGI = INT(T#): GGF = T# - GGI
     IF GGI > 2299160 THEN AGG = INT((GGI - 1867216.25#) / 36524.25#): BGG = GGI + 1 + AGG - INT(AGG / 4) ELSE BGG = GGI
     CGG = BGG + 1524: DGG = INT((CGG - 122.1) / 365.25): EGG = INT(365.25 * DGG): GGH = INT((CGG - EGG) / 30.6001#): GG = CGG - EGG + GGF - INT(30.6001 * GGH)
     IF GGH < 14 THEN MO% = GGH - 1 ELSE MO% = GGH - 13
     IF MO% > 2 THEN Y% = DGG - 4716 ELSE Y% = DGG - 4715
     D% = INT(GG): TH = (GG - D%) * 24: HR% = INT(TH): TMIN = (TH - HR%) * 60: MIN% = INT(TMIN): SEC% = (TMIN - MIN%) * 60
     IF SEC% = 60 THEN SEC% = 0: MIN% = MIN% + 1: IF MIN% = 60 THEN MIN% = 0: HR% = HR% + 1: IF HR% = 24 THEN HR% = 0: D% = D% + 1

END SUB

